# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

## Initialize SDPM
# Adrien Crovato

def init_sdpm(cfg):
    """Initialize SDPM components

    Parameters
    ----------
    cfg : dict
        Dictionary of parameters to configure SDPM

    Returns
    -------
    A dictionary containing
        n_f : int
            Number of reference reduced frequencies
        n_m : int
            Number of modes
        msh : sdpm.Mesh object
            Mesh data structure
        wrt : sdpm.GmshExport object
            Mesh writer
        pbl : sdpm.Probem object
            Problem definition
        bdy : sdpm.Body object
            Lifting body
        sol : sdpm.Solver object
            Direct solver
        grd : sdpm.Gradient object
            Gradient calculator
    """
    # Imports
    import sdpm
    from sdpm.gmsh import MeshLoader
    from sdpm.utils import interpolate_modes, interpolate_pressure
    import math

    # Sanity checks
    if 'AoS' in cfg and cfg['AoS'] != 0 and cfg['Symmetry']:
        raise RuntimeError('Symmetric (half) geometric model cannot be used with nonzero angle of sideslip (AoS)!\n')

    # Mesh
    pars = cfg.get('Pars', {})
    _msh = MeshLoader(cfg['File']).run(pars)
    _wrt = sdpm.GmshExport(_msh)

    # Problem
    aoa = cfg.get('AoA', 0.) * math.pi / 180
    aos = cfg.get('AoS', 0.) * math.pi / 180
    minf = cfg.get('Mach', 0.)
    _pbl = sdpm.Problem(_msh)
    _pbl.setGeometry(cfg['s_ref'], cfg['c_ref'], cfg['x_ref'], cfg['y_ref'], cfg['z_ref'], cfg['Symmetry'])
    _pbl.setFreestream(aoa, aos, minf)
    _pbl.setUnsteady(cfg['Num_modes'], cfg['Frequencies'])

    # Body
    _bdy = sdpm.Body(_msh, cfg['Wing'], cfg['Te'], cfg['c_ref'], cfg['Num_wake_div'], cfg['Num_modes'], len(cfg['Frequencies']))
    for i in range(cfg['Num_modes']):
        _bdy.addMotion()
    if 'Modes' in cfg:
        x_modes, amp_modes = interpolate_modes(cfg['Modes'], _bdy.getNodes())
        for i in range(cfg['Num_modes']):
            _bdy.setFlexibleMotion(i, 1 / amp_modes[i], x_modes[:, 3*i:3*i+3])
    if 'Transonic_pressure_grad' in cfg:
        dcp = interpolate_pressure(cfg['Transonic_pressure_grad'], _bdy.getNodes())
        _bdy.setTransonicPressureGrad(dcp[:, 0])
    _pbl.addBody(_bdy)

    # Solver
    vrb = cfg.get('Verb_lvl', 1)
    _sol = sdpm.Solver(_pbl, vrb) if 'Transonic_pressure_grad' not in cfg else sdpm.SolverTransonic(_pbl, vrb)

    # Gradient
    _grd = sdpm.Gradient(_sol)

    # Return
    return {
        'n_f': len(cfg['Frequencies']),
        'n_m': cfg['Num_modes'],
        'msh': _msh,
        'wrt': _wrt,
        'pbl': _pbl,
        'bdy': _bdy,
        'sol': _sol,
        'grd': _grd
    }
