# -*- coding: utf8 -*-
# gmsh utilities

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path

## Gmsh utilities
#
#  Adrien Crovato
#  Initally written by Romain Boman (https://gitlab.uliege.be/am-dept/amfe/-/blob/master/tbox/gmsh.py)
class MeshLoader:
    """Load a geometry (.geo) or mesh (.msh) file from Gmsh into SDPM data structure.
       Gmsh must be in the PATH.
    """
    def __init__(self, fname, vrb=False):
        # Check if given file exists and the set the filenames according to extension
        if os.path.isfile(fname):
            ext = os.path.splitext(fname)[1]
            if ext == '.geo':
                self.geofile = fname
                self.mshfile = None
            elif ext == '.msh':
                self.geofile = None
                self.mshfile = fname
            else:
                raise RuntimeError(f'{self.__class__}: bad gmsh extension ".{ext}"!\n')
        else:
            raise RuntimeError(f'{self.__class__}: file {fname} not found!\n')
        # Set the verbosity
        self.verb = vrb

    def run(self, pars={}):
        """Run Gmsh if necessary and load the mesh into SDPM data structure
        """
        if self.geofile is not None:
            self.__run_gmsh(pars)
        return self.__load_msh()

    def __print(self, string, end='\n'):
        """Print according to level of verbosity
        """
        if self.verb: print(f'{self.__class__}: ' + string, end=end)

    def __build_pars_list(self, pars):
        """Build the option and parameter lists to be passed to Gmsh
        """
        parcmd = ''
        for k,v in pars.items():
            if k[0] == '-':
                parcmd += f'{k} {v} ' # option
            else:
                parcmd += f'-setnumber {k} {v} ' # parameter
        return parcmd

    def __run_gmsh(self, pars):
        """Run Gmsh to generate the mesh if necessary
        """
        # Build the mesh and parameters filenames (not their path)
        self.mshfile = os.path.basename(self.geofile.replace('.geo', '.msh'))
        parfile = os.path.basename(self.geofile.replace('.geo', '.par'))
        # Get the new parameters
        parnew = str(pars)

        # Check if Gmsh needs to generate a new mesh
        # - if the mesh does not exist
        #   - if the .geo is newer than the old
        #   - if the parameters do not exist or have changed
        generate = True
        self.__print('checking if mesh needs to be generated...', end=' ')
        if os.path.isfile(self.mshfile):
            if  os.path.getmtime(self.geofile) > os.path.getmtime(self.mshfile):
                self.__print('yes, geometry file is newer than existing mesh file.')
            else:
                if os.path.isfile(parfile):
                    f = open(parfile, 'r')
                    parold = f.readline()
                    f.close()
                    if parold != parnew:
                        self.__print('yes, parameters have changed.')
                    else:
                        generate = False
                        self.__print('no, mesh file already exists and is up to date.')
                else:
                    self.__print('yes, parameters file does not exist.')
        else:
            self.__print('yes, mesh file does not exist.')

        # Call Gmsh to generate the mesh and write the parameters to file
        if generate:
            # Build the command line
            parcmd = self.__build_pars_list(pars)
            cmd = f'gmsh -3 -format msh2 {parcmd} -o "{self.mshfile}" "{self.geofile}"' # msh2 format because the sdpm::GmshImport can only read old format
            self.__print(f'running Gmsh on {self.geofile} using {cmd}')
            # Call Gmsh
            import shlex, subprocess
            try:
                sproc = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=os.environ)
                out, err = sproc.communicate()
            except Exception as e:
                print('Check that Gmsh in your PATH envrionment variable!\n')
                raise e
            # Redirect streams
            for l in out.splitlines():
                print('[gmsh-out] ', l)
            for l in err.splitlines():
                print('[gmsh-err] ', l)

            # Check that mesh file has been written
            if not os.path.isfile(self.mshfile):
                raise RuntimeError(f'{self.__class__}: mesh file {self.mshfile} does not exist although it should have been generated!\n')
            # Write parameters to file
            f = open(parfile,'w')
            f.write(parnew)
            f.close()

    def __load_msh(self):
        """Load the mesh into SDPM data structure
        """
        import sdpm
        msh = sdpm.Mesh()
        gmshLoader = sdpm.GmshImport(msh)
        gmshLoader.load(self.mshfile)
        return msh
