/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpm.h"
#include "sdpmBody.h"
#include "sdpmMesh.h"
#include "sdpmElement.h"
#include "sdpmQuad4.h"
#include "sdpmTag.h"
#include "sdpmGmshExport.h"
#include <unordered_set>
#include <iostream>
using namespace sdpm;

Body::Body(Mesh &msh, std::string const &name, std::string const &teName,
           double shedLgth, size_t nShedDiv,
           size_t nM, size_t nF) : Group(msh, name),
                                   _cl(0), _cd(0), _cs(0), _cm(0),
                                   _cl1Re(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cl1Im(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cd1Re(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cd1Im(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cs1Re(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cs1Im(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cm1Re(nM, std::vector<sdpmDouble>(nF, 0.)),
                                   _cm1Im(nM, std::vector<sdpmDouble>(nF, 0.))
{
    // Create wake
    std::cout << "Creating wake... " << std::flush;
    std::string wkName = _tag->getName() + "Wake";
    // get TE nodes and sort them according to their y-coordinate (ascending)
    Group te(_msh, teName);
    for (auto e : te.getElements())
        for (auto n : e->getNodes())
            _teNodes.push_back(n);
    std::sort(_teNodes.begin(), _teNodes.end(), [](Node *a, Node *b) -> bool
              { return a->getCoords()(1) < b->getCoords()(1); });
    _teNodes.erase(std::unique(_teNodes.begin(), _teNodes.end()), _teNodes.end());
    size_t nTe = _teNodes.size() - 1;
    // create wake geometry if it does not already exist
    std::map<std::string, Tag *> tags = _msh.getTags();
    auto it = tags.find(wkName);
    if (it == tags.end())
    {
        // sanity check
        if (shedLgth <= 0.)
        {
            std::stringstream err;
            err << "sdpm::Body: zero or negative characteristic chord length given for " << *_tag << "!\n";
            throw std::runtime_error(err.str());
        }
        // duplicate TE nodes
        std::map<Node *, Node *> teMap;
        for (auto n : _teNodes)
        {
            Node *nN = new Node(_msh.getNodes().back()->getId() + 1, n->getCoords());
            _msh.addNode(nN);
            teMap[n] = nN;
        }
        // create tag and add it to the mesh
        Tag *tagp = new Tag(tags.size() + 1, wkName, 2);
        _msh.addTag(wkName, tagp);
        // create wake nodes and elements
        std::vector<Node *> wkNodes = _teNodes;
        for (size_t i = 0; i < 10 * nShedDiv; ++i)
        {
            // translate TE nodes along x-coordinate
            sdpmVector3d delta((i + 1) * shedLgth / nShedDiv, 0., 0.);
            for (auto n : _teNodes)
            {
                Node *nN = new Node(_msh.getNodes().back()->getId() + 1, n->getCoords() + delta);
                wkNodes.push_back(nN);
                _msh.addNode(nN);
            }
            // create elements
            for (size_t j = 0; j < nTe; ++j)
            {
                std::vector<Node *> qnodes = {wkNodes[i * (nTe + 1) + j], wkNodes[(i + 1) * (nTe + 1) + j], wkNodes[(i + 1) * (nTe + 1) + j + 1], wkNodes[i * (nTe + 1) + j + 1]};
                _msh.addElement(new Quad4(_msh.getElements().back()->getId() + 1, tagp, qnodes));
            }
        }
        // create wake
        _wake = new Wake(_msh, wkName, _tag, nTe);
        // swap nodes
        std::unordered_set<Element *> lwEl; // get set of unique wing TE elements on the pressure side
        for (auto p : _wake->getElMap())
            lwEl.insert(p.second.second);
        for (auto e : lwEl)
        {
            std::vector<Node *> lwNods = e->getNodes();
            for (size_t i = 0; i < lwNods.size(); ++i)
            {
                try
                {
                    e->setNode(i, teMap.at(lwNods[i]));
                }
                catch (const std::out_of_range &)
                {
                    // std::cout << lwe->nodes[i]->no << "not found in map!\n";
                }
            }
        }
        // get body nodes (now that they are duplicated and swaped at lower TE)
        getUniqueNodes();
        // print and save new mesh
        std::cout << *_wake << "created. " << std::flush;
        GmshExport writer(_msh);
        writer.save();
    }
    else
    {
        // get body nodes
        getUniqueNodes();
        // map the lower TE nodes to their upper nodes
        std::map<Node *, Node *> iTeMap;
        for (auto ten : _teNodes)
            for (auto wn : _nodes)
                if (ten->getCoords() == wn->getCoords() && ten->getId() != wn->getId())
                {
                    iTeMap[wn] = ten;
                    break;
                }
        // create wake and print
        _wake = new Wake(_msh, wkName, _tag, nTe, iTeMap);
        std::cout << *_wake << "already existing, nothing done." << std::endl;
        // check if parameters match existing wake geometry
        std::vector<Node *> wNods = _wake->getNodes();
        if (std::abs(wNods[nTe + 1]->getCoords()(0) - wNods[0]->getCoords()(0) - shedLgth / nShedDiv) > 1e-12 || _wake->getElements().size() != 10 * nShedDiv * nTe)
        {
            std::stringstream err;
            err << "sdpm::Body: parameters \'shedLgth\' and \'nShedDiv\' do not match existing wake geometry!\n";
            throw std::runtime_error(err.str());
        }
    }

    // Associate each node to its adjacent elements
    for (auto e : _tag->getElements())
        for (auto n : e->getNodes())
            _neMap[n].push_back(e);

    // Size vectors
    _dGradPressure.resize(_msh.getNodes().size(), 0.);
    _nLoads.resize(_nodes.size(), sdpmVector3d::Zero());
    _nLoads1Re.resize(nM, std::vector<std::vector<sdpmVector3d>>(nF, std::vector<sdpmVector3d>(_nodes.size(), sdpmVector3d::Zero())));
    _nLoads1Im.resize(nM, std::vector<std::vector<sdpmVector3d>>(nF, std::vector<sdpmVector3d>(_nodes.size(), sdpmVector3d::Zero())));
    _iVelocityS.resize(nM, std::vector<sdpmVector3d>(getElements().size(), sdpmVector3d::Zero()));
    _iVelocityH.resize(nM, std::vector<sdpmVector3d>(getElements().size(), sdpmVector3d::Zero()));
}

Body::~Body()
{
    delete _wake;
    for (auto m : _motions)
        delete m;
    _motions.clear();
    std::cout << "~sdpm::Body()\n";
}

/**
 * @brief Set new coordinates for all nodes
 */
void Body::setNodes(std::vector<std::vector<double>> const &coords)
{
    // Check size
    if (coords.size() != _nodes.size())
        throw std::runtime_error("sdpm::Problem: size of coordinates vector must be equal to the number of nodes!\n");
    // Set nodes coordinates
    for (size_t i = 0; i < _nodes.size(); ++i)
        _nodes[i]->setCoords(sdpmVector3d(coords[i][0], coords[i][1], coords[i][2]));
    // Update wake
    sdpmDouble dx = _wake->getLagMap().at(_wake->getElements()[0]); // dx = dt of first cell
    size_t nTe = _teNodes.size();
    std::vector<Node *> wkNodes = _wake->getNodes();
    for (size_t i = 1; i < wkNodes.size() / nTe; ++i)
        for (size_t j = 0; j < nTe; ++j)
            wkNodes[i * nTe + j]->setCoords(_teNodes[j]->getCoords() + sdpmVector3d(i * dx, 0., 0.));
}

/**
 * @brief Add motion
 */
void Body::addMotion()
{
    _motions.push_back(new Motion(_msh.getNodes().size()));
}

/**
 * @brief Set rigid motion parameters
 */
void Body::setRigidMotion(size_t m, double aAmp, double hAmp)
{
    _motions[m]->setRigid(aAmp, hAmp);
}

/**
 * @brief Set flexible motion parameters
 */
void Body::setFlexibleMotion(size_t m, double mAmp, std::vector<std::vector<double>> const &xMod)
{
    // Check size
    if (xMod.size() != _nodes.size())
        throw std::runtime_error("sdpm::Problem: size of modal displacement vectors must be equal to the number of nodes!\n");
    _motions[m]->setFlexible(mAmp, xMod, _nodes);
}

/**
 * @brief Set transonic pressure derivative wrt to angle of attack on surface nodes
 */
void Body::setTransonicPressureGrad(std::vector<double> const &dCpA)
{
    for (size_t i = 0; i < _nodes.size(); ++i)
        _dGradPressure[_nodes[i]->getId() - 1] = dCpA[i];
}

/**
 * @brief Return the elements on the trailing edge on the suction and lower sides
 */
std::pair<std::vector<Element *>, std::vector<Element *>> Body::getTrailingEdgeElements() const
{
    std::vector<Element *> we = _wake->getElements();
    std::map<Element *, std::pair<Element *, Element *>> wwMap = _wake->getElMap();
    size_t nTe = _teNodes.size() - 1;
    std::vector<Element *> up(nTe), lw(nTe);
    for (size_t i = 0; i < nTe; ++i)
    {
        std::pair<Element *, Element *> p = wwMap.at(we[i]);
        up[i] = p.first;
        lw[i] = p.second;
    }
    return std::make_pair(up, lw);
}

/**
 * @brief Compute the motion-induced velocity on the body surface
 */
void Body::computeVelocity(size_t m, sdpmVector3d const &ui, sdpmDouble xRef, sdpmDouble zRef)
{
    std::vector<Element *> elems = getElements();
    std::vector<sdpmVector3d> ucS(elems.size()), ucH(elems.size());
    for (size_t i = 0; i < elems.size(); ++i)
    {
        ucS[i] = _motions[m]->computeSteady(*elems[i], ui);
        ucH[i] = _motions[m]->computeHarmonic(*elems[i], xRef, zRef);
    }
    _iVelocityS[m] = ucS;
    _iVelocityH[m] = ucH;
}

/**
 * @brief Return a vector of unique nodes belonging to the tag
 */
void Body::getUniqueNodes()
{
    for (auto e : _tag->getElements())
        for (auto n : e->getNodes())
            _nodes.push_back(n);
    std::sort(_nodes.begin(), _nodes.end());
    _nodes.erase(std::unique(_nodes.begin(), _nodes.end()), _nodes.end());
}

void Body::write(std::ostream &out) const
{
    out << "sdpm::Body " << *_tag << "with " << *_wake;
}
