/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMBODY_H
#define SDPMBODY_H

#include "sdpm.h"
#include "sdpmGroup.h"
#include "sdpmWake.h"
#include "sdpmMotion.h"
#include <map>

namespace sdpm
{

/**
 * @brief Manage a lifting body immersed in the fluid
 * @authors Adrien Crovato
 */
class SDPM_API Body : public Group
{
    // Geometry
    std::vector<Node *> _nodes;                      ///< nodes of the surface
    std::vector<Node *> _teNodes;                    ///< nodes of the (upper) trailing edge
    std::map<Node *, std::vector<Element *>> _neMap; ///< map between surface nodes and adjacent elements
    Wake *_wake;                                     ///< wake attached to the lifting body
    // Motion
    std::vector<Motion *> _motions;                     ///< body motions
    std::vector<std::vector<sdpmVector3d>> _iVelocityS; ///< steady motion induced velocity
    std::vector<std::vector<sdpmVector3d>> _iVelocityH; ///< harmonic motion induced velocity
    // Transonic correction data
    std::vector<sdpmDouble> _dGradPressure; ///< derivative of pressure coefficient wrt angle of attack
    // Nodal aerodynamic load (normalized by dynamic pressure)
    std::vector<sdpmVector3d> _nLoads;                              ///< steady nodal aerodynamic load
    std::vector<std::vector<std::vector<sdpmVector3d>>> _nLoads1Re; ///< real part of unsteady first harmonic nodal aerodynamic load
    std::vector<std::vector<std::vector<sdpmVector3d>>> _nLoads1Im; ///< imaginary part of unsteady first harmonic nodal aerodynamic load
    //  Coefficients
    sdpmDouble _cl;                              ///< steady lift coefficient
    sdpmDouble _cd;                              ///< steady drag coefficient
    sdpmDouble _cs;                              ///< steady side force coefficient
    sdpmDouble _cm;                              ///< steady pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmDouble>> _cl1Re; ///< real part of unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmDouble>> _cl1Im; ///< imaginary part of unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmDouble>> _cd1Re; ///< real part of unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmDouble>> _cd1Im; ///< imaginary part of unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmDouble>> _cs1Re; ///< real part of unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmDouble>> _cs1Im; ///< imaginary part of unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmDouble>> _cm1Re; ///< real part of unsteady first harmonic pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmDouble>> _cm1Im; ///< imaginary part of unsteady first harmonic pitch moment coefficient (positive nose-up)

public:
    Body(Mesh &msh, std::string const &name, std::string const &teName, double shedLgth, size_t nShedDiv = 1, size_t nM = 0, size_t nF = 0);
    ~Body();

private:
    void getUniqueNodes();

public:
    std::vector<Node *> const &getNodes() const { return _nodes; }
    std::vector<sdpmVector3d> const &getLoads() const { return _nLoads; }
    std::vector<sdpmVector3d> const &getUnsteadyLoadsReal(size_t imd, size_t ifq) const { return _nLoads1Re[imd][ifq]; }
    std::vector<sdpmVector3d> const &getUnsteadyLoadsImag(size_t imd, size_t ifq) const { return _nLoads1Im[imd][ifq]; }
    sdpmDouble getLiftCoeff() const { return _cl; }
    sdpmDouble getDragCoeff() const { return _cd; }
    sdpmDouble getSideforceCoeff() const { return _cs; }
    sdpmDouble getMomentCoeff() const { return _cm; }
    sdpmDouble getUnsteadyLiftCoeffReal(size_t imd, size_t ifq) const { return _cl1Re[imd][ifq]; }
    sdpmDouble getUnsteadyLiftCoeffImag(size_t imd, size_t ifq) const { return _cl1Im[imd][ifq]; }
    sdpmDouble getUnsteadyDragCoeffReal(size_t imd, size_t ifq) const { return _cd1Re[imd][ifq]; }
    sdpmDouble getUnsteadyDragCoeffImag(size_t imd, size_t ifq) const { return _cd1Im[imd][ifq]; }
    sdpmDouble getUnsteadySideforceCoeffReal(size_t imd, size_t ifq) const { return _cs1Re[imd][ifq]; }
    sdpmDouble getUnsteadySideforceCoeffImag(size_t imd, size_t ifq) const { return _cs1Im[imd][ifq]; }
    sdpmDouble getUnsteadyMomentCoeffReal(size_t imd, size_t ifq) const { return _cm1Re[imd][ifq]; }
    sdpmDouble getUnsteadyMomentCoeffImag(size_t imd, size_t ifq) const { return _cm1Im[imd][ifq]; }
    void setNodes(std::vector<std::vector<double>> const &coords);
    void addMotion();
    void setRigidMotion(size_t m, double aAmp, double hAmp);
    void setFlexibleMotion(size_t m, double mAmp, std::vector<std::vector<double>> const &xMod);
    void setTransonicPressureGrad(std::vector<double> const &dCpA);
#ifndef SWIG
    std::pair<std::vector<Element *>, std::vector<Element *>> getTrailingEdgeElements() const;
    std::map<Node *, std::vector<Element *>> const &getMap() const { return _neMap; }
    Wake const &getWake() const { return *_wake; }
    size_t getNumMotions() const { return _motions.size(); }
    sdpmDouble getPitchAmplitude(size_t imd) const { return _motions[imd]->getPitchAmplitude(); }
    sdpmDouble getPlungeAmplitude(size_t imd) const { return _motions[imd]->getPlungeAmplitude(); }
    std::vector<sdpmDouble> const &getModeZ(size_t imd) const { return _motions[imd]->getModeZ(); }
    std::vector<sdpmVector3d> const &getSteadyVelocity(size_t imd) const { return _iVelocityS[imd]; };
    std::vector<sdpmVector3d> const &getHarmonicVelocity(size_t imd) const { return _iVelocityH[imd]; };
    std::vector<sdpmDouble> const &getTransonicPressureGrad() const { return _dGradPressure; }
    void setLoads(std::vector<sdpmVector3d> const &nl) { _nLoads = nl; }
    inline void setUnsteadyLoads(size_t imd, size_t ifq, std::vector<sdpmVector3cd> const &nl);
    void setLiftCoeff(sdpmDouble cl) { _cl = cl; }
    void setDragCoeff(sdpmDouble cd) { _cd = cd; }
    void setSideforceCoeff(sdpmDouble cs) { _cs = cs; }
    void setMomentCoeff(sdpmDouble cm) { _cm = cm; }
    inline void setUnsteadyLiftCoeff(size_t imd, size_t ifq, sdpmComplex cl);
    inline void setUnsteadyDragCoeff(size_t imd, size_t ifq, sdpmComplex cd);
    inline void setUnsteadySideforceCoeff(size_t imd, size_t ifq, sdpmComplex cs);
    inline void setUnsteadyMomentCoeff(size_t imd, size_t ifq, sdpmComplex cm);
    void computeVelocity(size_t imd, sdpmVector3d const &ui, sdpmDouble xRef, sdpmDouble zRef);
    virtual void write(std::ostream &out) const override;
#endif
};

#ifndef SWIG
#include "sdpmBody.inl"
#endif

} // namespace sdpm

#endif // SDPMBODY_H
