/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmElement.h"

using namespace sdpm;

namespace sdpm
{
SDPM_API std::ostream &operator<<(std::ostream &out, ElType const &obj)
{
    switch (obj)
    {
    case ElType::LINE2:
        out << "LINE2";
        break;
    case ElType::QUAD4:
        out << "QUAD4";
        break;
    default:
        out << "UNKNOWN-" << static_cast<int>(obj);
        break;
    }
    return out;
}
} // namespace sdpm

Element::Element(size_t n, Tag *tag, std::vector<Node *> &nodes) : sdpmObject(), _id(n), _tag(tag), _nodes(nodes), _toDate(false)
{
    _coords.resize(_nodes.size());
    _tag->addElements(this);
}

void Element::update()
{
    throw std::runtime_error("Element::update not implemented\n");
}

/**
 * @brief Update compressible coordinates and center of gravity
 */
void Element::update(sdpmDouble beta)
{
    // Scale nodes and compute center of gravity
    _cg = sdpmVector3d::Zero();
    for (size_t i = 0; i < _nodes.size(); ++i)
    {
        _coords[i] = _nodes[i]->getCoords().cwiseQuotient(sdpmVector3d(beta, 1., 1.));
        _cg += _coords[i];
    }
    _cg /= static_cast<sdpmDouble>(_nodes.size());
}

sdpmMatrixXd Element::computeGradientOperator() const
{
    throw std::runtime_error("Element::computeGradientOperator not implemented\n");
}

sdpmVector3d Element::computeGradient(std::vector<sdpmDouble> const &u) const
{
    throw std::runtime_error("Element::computeGradient not implemented\n");
}

void Element::write(std::ostream &out) const
{
    // id
    out << getType() << " #" << _id;
    // tag
    out << " in tag #" << _tag->getId() << " \"" << _tag->getName() << "\"";
    // nodes
    out << " with nodes=[ ";
    for (auto n : _nodes)
        out << "#" << n->getId() << ' ';
    out << "]";
}
