/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmGmshImport.h"
#include "sdpmTag.h"
#include "sdpmElement.h"
#include "sdpmNode.h"
#include "sdpmLine2.h"
#include "sdpmQuad4.h"
#include <iostream>

using namespace sdpm;

GmshImport::GmshImport(Mesh &msh, bool v) : sdpmObject(), _msh(msh), _verb(v)
{
}

GmshImport::~GmshImport()
{
    _nodeMap.clear();
    _tagMap.clear();
    std::cout << "~sdpm::GmshImport()\n";
}

/**
 * @brief Search for a line in a code
 */
int GmshImport::searchCode(char const *code, char const *line) const
{
    return strncmp(code, line, strlen(code));
}

/**
 * @brief Read mesh format type (version)
 */
void GmshImport::readMeshFormat(FILE *file)
{
    double version = 0.0;
    int ftype = 0, dsize = 0;
    if (fscanf(file, "%lf %d %d\n", &version, &ftype, &dsize) != 3)
        throw std::runtime_error("GmshImport::readMeshFormat: error reading file!\n");
    if (ftype != 0)
        throw std::runtime_error("GmshImport::readMeshFormat: cannot read binary format!\n");
    double eps = 1e-5;
    if (version < 2.2 - eps || version > 2.2 + eps)
    {
        std::stringstream str;
        str << "GmshImport::readMeshFormat: bad .msh version: (" << version << ")!\n";
        throw std::runtime_error(str.str());
    }
    else
        std::cout << "\treading .msh format " << version << "\n";
}

/**
 * @brief Read physical group names
 */
void GmshImport::readPhysicalTags(FILE *file)
{
    // read number of names
    int nbtag = 0;
    if (fscanf(file, "%d\n", &nbtag) != 1)
        throw std::runtime_error("GmshImport::readPhysicalTags: error reading number of physical names!\n");

    // read the names
    for (int i = 0; i < nbtag; ++i)
    {
        // read dimension, label and name
        size_t dim = 0, no = 0;
        char name[128];
        if (fscanf(file, "%zu %zu \"%[^\"]\"", &dim, &no, name) != 3)
            throw std::runtime_error("GmshImport::readPhysicalTags: error reading dim, no, name!\n");
        // add tag to tag list
        std::map<std::string, Tag *> tags = _msh.getTags();
        auto it = tags.find(name);
        if (it == tags.end())
        {
            Tag *tag = new Tag(no, name, dim);
            _tagMap[no] = tag;
            _msh.addTag(name, tag);
        }
    }
}

/**
 * @brief Read the nodes
 */
void GmshImport::readNodes(FILE *file)
{
    // read number of nodes
    int nbnod = 0;
    if (fscanf(file, "%d\n", &nbnod) != 1)
        throw std::runtime_error("GmshImport::readNodes: error reading number of nodes!\n");
    std::cout << "\tcounted " << nbnod << " nodes\n";

    // allocate memory and read the nodes
    _msh.reserveNodes(nbnod);
    for (int i = 0; i < nbnod; ++i)
    {
        size_t no = 0;
        double x = 0.0, y = 0.0, z = 0.0;
        if (fscanf(file, "%zu %lf %lf %lf\n", &no, &x, &y, &z) != 4)
            break;
        if (_verb)
            std::cout << "\t" << no << ((i == nbnod - 1) ? ".\n" : ",");
        Node *nod = new Node(no, sdpmVector3d(x, y, z));
        _msh.addNode(nod);
        _nodeMap[no] = nod;
    }
}

/**
 * @brief Read the elements
 */
void GmshImport::readElements(FILE *file)
{
    // read number of elements
    int nbelm = 0;
    if (fscanf(file, "%d\n", &nbelm) != 1)
        throw std::runtime_error("GmshImport::readElements: error reading number of elements!\n");
    std::cout << "\tcounted " << nbelm << " elements\n";

    // allocate memory and read the elements
    _msh.reserveElements(nbelm);
    for (int i = 0; i < nbelm; ++i)
    {
        // read number, type and nb of tags
        size_t no = 0;
        int type = 0, nbtags = 0;
        if (fscanf(file, "%zu %d %d", &no, &type, &nbtags) != 3)
            throw std::runtime_error("GmshImport::readElements: error reading id, type, nbtags!\n");

        // read physical tag and skip the other kind of tags
        Tag *ptag = nullptr;
        if (nbtags > 0)
            ptag = findTag(file);
        if (nbtags > 1)
        {
            size_t idx;
            for (int itag = 1; itag < nbtags; ++itag)
                if (fscanf(file, "%zu", &idx) != 1)
                    throw std::runtime_error("GmshImport::readElements: error reading tags!\n");
        }

        // read/link nodes
        bool skipline = false;
        std::vector<Node *> nods;
        if (static_cast<ElType>(type) == ElType::QUAD4) // 4-node quad
        {
            readElNods(file, 4, nods);
            _msh.addElement(new Quad4(no, ptag, nods));
        }
        else if (static_cast<ElType>(type) == ElType::LINE2) // 2-node line
        {
            readElNods(file, 2, nods);
            _msh.addElement(new Line2(no, ptag, nods));
        }
        else
            skipline = true;

        if (skipline)
        {
            // skip line
            char line[128];
            if (fgets(line, sizeof line, file) == NULL)
                throw std::runtime_error("GmshImport::readElements: error skipping line!\n");
            if (_verb)
                std::cout << "[elm] skipping line:\"" << line << "\"\n";
        }
    }
}

/**
 * @brief Read nodes of an element
 */
void GmshImport::readElNods(FILE *file, int n, std::vector<Node *> &nods)
{
    int nn;
    for (int j = 0; j < n; ++j)
    {
        if (fscanf(file, "%d", &nn) != 1)
            throw std::runtime_error("GmshImport::readElNods: error reading element/node links!\n");

        auto it = _nodeMap.find(nn);
        if (it == _nodeMap.end())
            throw std::runtime_error("GmshImport::readElNods: unknown node label!\n");
        nods.push_back(it->second);
    }
}

/**
 * @brief Create a tag for each physical group
 */
Tag *GmshImport::findTag(FILE *file)
{
    size_t tagId;
    if (fscanf(file, "%zu", &tagId) != 1)
        throw std::runtime_error("GmshImport::findTag: error reading tags!\n");
    try
    {
        return _tagMap.at(tagId);
    }
    catch (std::out_of_range &)
    {
        throw std::runtime_error("GmshImport::findTag: tag number tagId does not exist!\n");
    }
}

/**
 * @brief Load a mesh from a file
 */
void GmshImport::load(std::string const &fname)
{
    // clear data structure
    _msh.clear();
    _nodeMap.clear();
    _tagMap.clear();

    // read filename
    std::cout << "reading file " << fname << "...\n";
    auto idx = fname.find(".msh");
    if (idx != std::string::npos)
    {
        _msh.setName(fname.substr(0, idx));
        std::cout << "\tsetting mesh name to \"" << _msh.getName() << "\"\n";
    }

    // read gmsh .msh file with stdio commands
    FILE *file = fopen(fname.c_str(), "r");
    if (!file)
        throw std::runtime_error("GmshImport::load: error opening file!\n");

    int skipped = 0;
    while (1)
    {
        // read next line
        char line[128];
        if (fgets(line, sizeof line, file) == NULL)
            break;

        // process code
        if (searchCode("$MeshFormat", line) == 0)
            readMeshFormat(file);
        else if (searchCode("$PhysicalNames", line) == 0)
            readPhysicalTags(file);
        else if (searchCode("$Nodes", line) == 0)
            readNodes(file);
        else if (searchCode("$Elements", line) == 0)
            readElements(file);
        else
        {
            skipped += 1;
            if (_verb)
                std::cout << "\t"
                          << "[code] skipping line:\"" << line << "\"\n";
        }
    }
    if (_verb)
        std::cout << skipped << " lines skipped\n";

    // close file
    fclose(file);
    _nodeMap.clear();
    _tagMap.clear();
    std::cout << "done." << std::endl;
}

void GmshImport::write(std::ostream &out) const
{
    out << "GmshImport on " << _msh << std::endl;
}
