/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmProblem.h"
#include "sdpmBody.h"
#include "sdpmElement.h"
#include "sdpmTag.h"
#include <iostream>
using namespace sdpm;

Problem::Problem(Mesh &msh) : _msh(msh),
                              _aoa(0.), _aos(0.), _mach(0.), _beta(1.0),
                              _dDir(sdpmVector3d::Zero()), _sDir(sdpmVector3d::Zero()), _lDir(sdpmVector3d::Zero()),
                              _sRef(0.), _cRef(0.), _xRef(sdpmVector3d::Zero()), _ySym(false),
                              _nModes(0),
                              _hasAero(false), _hasGeo(false)
{
}

Problem::~Problem()
{
    std::cout << "~sdpm::Problem()\n";
}

/**
 * @brief Add a non-lifting body
 */
void Problem::addBody(Body &b)
{
    _bodies.push_back(&b);
}

/**
 * @brief Set the freestream aerodynamic values
 */
void Problem::setFreestream(double aoa, double aos, double mch)
{
    // Angles
    _aoa = aoa;
    _aos = aos;
    // Mach number
    _mach = mch;
    // Flag
    _hasAero = true;
}

/**
 * @brief Set the reference geometric values
 */
void Problem::setGeometry(double sref, double cref, double xref, double yref, double zref, bool ysym)
{
    // Reference values
    _sRef = sref;
    _cRef = cref;
    _xRef = sdpmVector3d(xref, yref, zref);
    _ySym = ysym;
    // Flag
    _hasGeo = true;
}

/**
 * @brief Set the unsteady number of modes and frequencies
 */
void Problem::setUnsteady(size_t nmods, std::vector<double> const &ks)
{
    // Check if chord length is available
    if (!_hasGeo)
        throw std::runtime_error("sdpm::Problem: geometric parameters have not been provided!\n");
    // Set parameters
    _nModes = nmods;
    _omega.clear();
    for (size_t i = 0; i < ks.size(); ++i)
        _omega.push_back(ks[i] / (0.5 * _cRef));
}

/**
 * @brief Update the problem
 */
void Problem::update()
{
    // Compressibility
    _beta = std::sqrt(1 - _mach * _mach);
    // Directions
    _dDir = sdpmVector3d(cos(_aoa) * cos(_aos), sin(_aos), sin(_aoa) * cos(_aos));
    _sDir = sdpmVector3d(-cos(_aoa) * sin(_aos), cos(_aos), -sin(_aoa) * sin(_aos));
    _lDir = sdpmVector3d(-sin(_aoa), 0., cos(_aoa));
    // Unscaled and scaled geometry
    for (auto b : _bodies)
    {
        for (auto e : b->getElements())
        {
            e->update();
            e->update(_beta);
        }
        for (auto e : b->getWake().getElements())
        {
            e->update();
            e->update(_beta);
        }
    }
    // Unsteady motion-induced velocities
    for (size_t imd = 0; imd < _nModes; ++imd)
        for (auto body : _bodies)
            body->computeVelocity(imd, _dDir, _xRef(0), _xRef(2));
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (!_hasAero)
        throw std::runtime_error("sdpm::Problem: aerodynamic parameters have not been provided!\n");
    if (!_hasGeo)
        throw std::runtime_error("sdpm::Problem: geometric parameters have not been provided!\n");
    if (_bodies.empty())
        throw std::runtime_error("sdpm::Problem: no lifting bodies provided!\n");
    for (auto b : _bodies)
        if (b->getNumMotions() != _nModes)
            throw std::runtime_error("sdpm::Problem: Body and Problem do not have the same number of modes!\n");

    // Surface elements
    for (auto b : _bodies)
        for (auto e : b->getElements())
            if (e->getType() != ElType::QUAD4)
            {
                std::stringstream err;
                err << "SDPM is only implemented for surface elements of type "
                    << ElType::QUAD4 << " (" << e->getType() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
}

void Problem::write(std::ostream &out) const
{
    out << "sdpm::Problem definition"
        << "\n\tAngle of attack: " << _aoa
        << "\n\tAngle of sideslip: " << _aos
        << "\n\tMach number: " << _mach
        << "\n\tReference area: " << _sRef
        << "\n\tReference length: " << _cRef
        << "\n\tReference point: [" << _xRef.transpose() << "]"
        << std::endl;
    out << "with";
    for (auto b : _bodies)
        out << "\t" << *b;
}
