/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmSolver.h"
#include "sdpmBuilder.h"
#include "sdpmProblem.h"
#include "sdpmBody.h"
#include "sdpmMesh.h"
#include "sdpmGmshExport.h"
#include "sdpmResults.h"
#include "sdpm_ad_interface.h"
#include "sdpm_extras.h"
#include <iostream>
#include <iomanip>
#include <fstream>

using namespace sdpm;

Solver::Solver(Problem &pbl, int vrb) : _pbl(pbl), _vrb(vrb), _cl(0), _cd(0), _cs(0), _cm(0)
{
    // Say hi
    std::cout << std::endl;
    std::cout << "*******************************************************************************" << std::endl;
    std::cout << "** Hi! My name is SDPM v1.2-2412                                             **" << std::endl;
    std::cout << "** ULiege 2023-2024                                                          **" << std::endl;
    std::cout << "*******************************************************************************" << std::endl;

    // Check the problem
    _pbl.check();

    // Count panel and set up global->local id map
    _nP = 0;
    int i = 0;
    for (auto body : _pbl.getBodies())
    {
        _nP += body->getElements().size();
        for (auto e : body->getElements())
            _rows[e] = i++;
    }

    // Get sizes
    size_t nm = _pbl.getNumModes();
    size_t nf = _pbl.getFrequencies().size();
    size_t ne = _pbl.getElements().size();

    // Resize AIC matrics
    _A0 = sdpmMatrixXd(_nP, _nP);
    _B0 = sdpmMatrixXd(_nP, _nP);
    _A.resize(nf, sdpmMatrixXcd::Zero(_nP, _nP));
    _B.resize(nf, sdpmMatrixXcd::Zero(_nP, _nP));

    // Set up variables
    _tau.resize(ne, 0.);
    _mu.resize(ne, 0.);
    _u.resize(ne, sdpmVector3d::Zero());
    _cp.resize(ne, 0.);
    _u1Re.resize(nm, std::vector<std::vector<sdpmVector3d>>(nf, std::vector<sdpmVector3d>(ne, sdpmVector3d::Zero())));
    _u1Im.resize(nm, std::vector<std::vector<sdpmVector3d>>(nf, std::vector<sdpmVector3d>(ne, sdpmVector3d::Zero())));
    _cp1Re.resize(nm, std::vector<std::vector<sdpmDouble>>(nf, std::vector<sdpmDouble>(ne, 0.)));
    _cp1Im.resize(nm, std::vector<std::vector<sdpmDouble>>(nf, std::vector<sdpmDouble>(ne, 0.)));
    _cl1Re.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cl1Im.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cd1Re.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cd1Im.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cs1Re.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cs1Im.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cm1Re.resize(nm, std::vector<sdpmDouble>(nf, 0.));
    _cm1Im.resize(nm, std::vector<sdpmDouble>(nf, 0.));

    // Resize GAF matrices
    _q1Re.resize(nf, sdpmMatrixXd::Zero(nm, nm));
    _q1Im.resize(nf, sdpmMatrixXd::Zero(nm, nm));

    // Display configuration
    std::cout << "--- Physical groups ---\n";
    std::cout << "Number of panels: " << _nP << "\n";
    for (auto b : _pbl.getBodies())
        std::cout << *b << "\n";
    std::cout << "--- Freestream conditions ---\n"
              << std::setprecision(5)
              << "Angle of attack: " << _pbl.getAoa() * 180 / 3.14159 << " deg\n"
              << "Angle of sideslip: " << _pbl.getAos() * 180 / 3.14159 << " deg\n"
              << "Mach number: " << _pbl.getMach() << "\n";
    std::cout << "--- Reference data ---\n"
              << "Surface area: " << _pbl.getRefArea() << "\n"
              << "Chord length: " << _pbl.getRefLgth() << "\n"
              << "Origin: (" << _pbl.getRefCtr().transpose() << ")\n";
    std::cout << "--- Unsteady parameters ---\n"
              << "Number of frequencies: " << _pbl.getFrequencies().size() << "\n"
              << "Number of motions: " << _pbl.getNumModes() << "\n"
              << std::endl;
}

Solver::~Solver()
{
    std::cout << "~sdpm::Solver()\n";
}

/**
 * @brief Update elements and AIC matrices
 */
void Solver::update()
{
    // Update and get values
    _pbl.update();
    std::vector<sdpmDouble> omega = _pbl.getFrequencies();
    sdpmDouble mi = _pbl.getMach();
    sdpmDouble beta = _pbl.getCompressibility();

    // Compute AIC
    std::cout << "Computing AIC matrices... " << std::flush;
    std::vector<Body *> bodies = _pbl.getBodies();
    _A0.setZero();
    _B0.setZero();
    std::fill(_A.begin(), _A.end(), sdpmMatrixXcd::Zero(_nP, _nP));
    std::fill(_B.begin(), _B.end(), sdpmMatrixXcd::Zero(_nP, _nP));
    for (auto jbody : bodies)
    {
        // body
        for (auto ej : jbody->getElements())
        {
            sdpmDouble nx = ej->getCompressibleNormal()(0);                // x-component of normal vector
            sdpmMatrix3d g2l = Builder::computeTransfoMat(false, ej);      // global to local transformation matrix
            sdpmMatrixXd sCoords = g2l * Builder::gatherCoords(false, ej); // coordinates of source panel in local axes
            for (auto ibody : bodies)
            {
                for (auto ei : ibody->getElements())
                {
                    std::vector<sdpmDouble> mutau = Builder::computeAIC(false, ei, ej, sCoords, g2l); // steady AIC
                    _A0(_rows.at(ei), _rows.at(ej)) = mutau[0];
                    _B0(_rows.at(ei), _rows.at(ej)) = mutau[1];
                    for (size_t fk = 0; fk < omega.size(); ++fk)
                    {
                        sdpmDouble om = omega[fk] * mi / beta; // u_inf~1
                        sdpmVector3d r = Builder::computeDistVec(false, ei, ej);
                        sdpmComplex exp = std::exp(sdpmComplex(0, -om * (r.norm() - mi * r(0))));
                        _A[fk](_rows.at(ei), _rows.at(ej)) = sdpmComplex(mutau[0], mutau[0] * om * r.norm() - mutau[1] * om * mi * nx) * exp;
                        _B[fk](_rows.at(ei), _rows.at(ej)) = mutau[1] * exp;
                    }
                }
            }
        }
        // wake
        Wake const &wake = jbody->getWake();
        std::map<Element *, std::pair<Element *, Element *>> wwMap = wake.getElMap();
        std::map<Element *, sdpmDouble> lagMap = wake.getLagMap();
        for (auto ew : wake.getElements())
        {
            sdpmMatrix3d g2l = Builder::computeTransfoMat(false, ew);      // global to local transformation matrix
            sdpmMatrixXd sCoords = g2l * Builder::gatherCoords(false, ew); // coordinates of source panel in local axes
            for (auto ibody : bodies)
            {
                for (auto ei : ibody->getElements())
                {
                    std::vector<sdpmDouble> mutau = Builder::computeAIC(false, ei, ew, sCoords, g2l); // steady AIC
                    _A0(_rows.at(ei), _rows.at(wwMap.at(ew).first)) += mutau[0];
                    _A0(_rows.at(ei), _rows.at(wwMap.at(ew).second)) -= mutau[0];
                    for (size_t fk = 0; fk < omega.size(); ++fk)
                    {
                        sdpmDouble om = omega[fk] * mi / beta; // u_inf~1
                        sdpmVector3d r = Builder::computeDistVec(false, ei, ew);
                        sdpmComplex exp = std::exp(sdpmComplex(0, -om * (r.norm() - mi * r(0))));
                        sdpmComplex waktrm = mutau[0] * sdpmComplex(1, om * r.norm()) * exp * std::exp(sdpmComplex(0, -omega[fk] * lagMap.at(ew)));
                        _A[fk](_rows.at(ei), _rows.at(wwMap.at(ew).first)) += waktrm;
                        _A[fk](_rows.at(ei), _rows.at(wwMap.at(ew).second)) -= waktrm;
                    }
                }
            }
        }
    }
    // symmetry
    if (_pbl.getSym())
    {
        for (auto jbody : bodies)
        {
            // body
            for (auto ej : jbody->getElements())
            {
                sdpmDouble nx = ej->getCompressibleNormal()(0);               // x-component of normal vector
                sdpmMatrix3d g2l = Builder::computeTransfoMat(true, ej);      // global to local transformation matrix
                sdpmMatrixXd sCoords = g2l * Builder::gatherCoords(true, ej); // coordinates of source panel in local axes
                for (auto ibody : bodies)
                {
                    for (auto ei : ibody->getElements())
                    {
                        std::vector<sdpmDouble> mutau = Builder::computeAIC(true, ei, ej, sCoords, g2l); // steady AIC
                        _A0(_rows.at(ei), _rows.at(ej)) -= mutau[0];
                        _B0(_rows.at(ei), _rows.at(ej)) -= mutau[1];
                        for (size_t fk = 0; fk < omega.size(); ++fk)
                        {
                            sdpmDouble om = omega[fk] * mi / beta; // u_inf~1
                            sdpmVector3d r = Builder::computeDistVec(true, ei, ej);
                            sdpmComplex exp = std::exp(sdpmComplex(0, -om * (r.norm() - mi * r(0))));
                            _A[fk](_rows.at(ei), _rows.at(ej)) -= sdpmComplex(mutau[0], mutau[0] * om * r.norm() - mutau[1] * om * mi * nx) * exp;
                            _B[fk](_rows.at(ei), _rows.at(ej)) -= mutau[1] * exp;
                        }
                    }
                }
            }
            // wake
            Wake const &wake = jbody->getWake();
            std::map<Element *, std::pair<Element *, Element *>> wwMap = wake.getElMap();
            std::map<Element *, sdpmDouble> lagMap = wake.getLagMap();
            for (auto ew : wake.getElements())
            {
                sdpmMatrix3d g2l = Builder::computeTransfoMat(true, ew);      // global to local transformation matrix
                sdpmMatrixXd sCoords = g2l * Builder::gatherCoords(true, ew); // coordinates of source panel in local axes
                for (auto ibody : bodies)
                {
                    for (auto ei : ibody->getElements())
                    {
                        std::vector<sdpmDouble> mutau = Builder::computeAIC(true, ei, ew, sCoords, g2l); // steady AIC
                        _A0(_rows.at(ei), _rows.at(wwMap.at(ew).first)) -= mutau[0];
                        _A0(_rows.at(ei), _rows.at(wwMap.at(ew).second)) += mutau[0];
                        for (size_t fk = 0; fk < omega.size(); ++fk)
                        {
                            sdpmDouble om = omega[fk] * mi / beta; // u_inf~1
                            sdpmVector3d r = Builder::computeDistVec(true, ei, ew);
                            sdpmComplex exp = std::exp(sdpmComplex(0, -om * (r.norm() - mi * r(0))));
                            sdpmComplex waktrm = mutau[0] * sdpmComplex(1, om * r.norm()) * exp * std::exp(sdpmComplex(0, -omega[fk] * lagMap.at(ew)));
                            _A[fk](_rows.at(ei), _rows.at(wwMap.at(ew).first)) -= waktrm;
                            _A[fk](_rows.at(ei), _rows.at(wwMap.at(ew).second)) += waktrm;
                        }
                    }
                }
            }
        }
    }
    std::cout << "done." << std::endl;
}

/**
 * @brief Solve the steady and unsteady potential equations
 */
void Solver::run()
{
    std::cout << "Running SDPM solver..." << std::flush;
    if (_vrb > 0)
        std::cout << "\n-- Mean flow\n"
                  << "computing sources... " << std::flush;
    sdpmVector3d ui = _pbl.getDragDir(); // freestream velocity
    sdpmVectorXd etau(_nP);
    for (auto body : _pbl.getBodies())
        for (auto e : body->getElements())
            etau(_rows[e]) = ui.dot(e->getCompressibleNormal().cwiseQuotient(sdpmVector3d(_pbl.getCompressibility(), 1., 1.)));
    if (_vrb > 0)
        std::cout << "done.\n"
                  << "solving for doublets... " << std::flush;
    sdpmVectorXd rhs = _B0 * etau;
    sdpmVectorXd emu = solve(_A0, rhs);
    if (_vrb > 0)
        std::cout << "done.\n"
                  << "computing flow and loads on bodies... " << std::flush;
    post(etau, emu);
    if (_vrb > 0)
        std::cout << "done." << std::endl;

    for (size_t imd = 0; imd < _pbl.getNumModes(); ++imd)
    {
        for (size_t ifq = 0; ifq < _pbl.getFrequencies().size(); ++ifq)
        {
            if (_vrb > 0)
                std::cout << "-- Mode #" << imd << ", frequency #" << ifq << "\n"
                          << "computing sources... " << std::flush;
            sdpmVectorXcd etau1(_nP);
            for (auto body : _pbl.getBodies())
            {
                std::vector<sdpmVector3d> uc0 = body->getSteadyVelocity(imd);
                std::vector<sdpmVector3d> uc1 = body->getHarmonicVelocity(imd);
                std::vector<Element *> elems = body->getElements();
                for (size_t i = 0; i < elems.size(); ++i)
                    etau1(_rows[elems[i]]) = (uc0[i] + uc1[i] * sdpmComplex(0., 1.) * _pbl.getFrequencies()[ifq]).conjugate().dot(elems[i]->getCompressibleNormal().cwiseQuotient(sdpmVector3d(_pbl.getCompressibility(), 1., 1.)));
            }
            if (_vrb > 0)
                std::cout << "done.\n"
                          << "solving for doublets... " << std::flush;
            // Split complex set of equations into real set (twice the size)
            sdpmMatrixXd A(2 * _nP, 2 * _nP);
            A.block(0, 0, _nP, _nP) = _A[ifq].real();
            A.block(_nP, _nP, _nP, _nP) = _A[ifq].real();
            A.block(_nP, 0, _nP, _nP) = _A[ifq].imag();
            A.block(0, _nP, _nP, _nP) = -_A[ifq].imag();
            sdpmVectorXd rhs1(2 * _nP);
            rhs1.block(0, 0, _nP, 1) = (_B[ifq] * etau1).real();
            rhs1.block(_nP, 0, _nP, 1) = (_B[ifq] * etau1).imag();
            sdpmVectorXd mu = solve(A, rhs1);
            sdpmVectorXcd emu1 = mu.block(0, 0, _nP, 1) + sdpmComplex(0, 1) * mu.block(_nP, 0, _nP, 1);
            if (_vrb > 0)
                std::cout << "done.\n"
                          << "computing flow and loads on bodies... " << std::flush;
            post(imd, ifq, etau1, emu1);
            if (_vrb > 0)
                std::cout << "done." << std::endl;
        }
    }
    if (_vrb > 0)
        std::cout << "-- Generalized Aerodynamic Force matrices" << std::endl;
    for (size_t ifq = 0; ifq < _pbl.getFrequencies().size(); ++ifq)
    {
        if (_vrb > 0)
            std::cout << "computing GAF at frequency #" << ifq << "... " << std::flush;
        computeGAF(ifq);
        if (_vrb > 0)
            std::cout << "done." << std::endl;
    }
    if (_vrb == 0)
        std::cout << " done.";
    std::cout << std::endl;
}

/**
 * @brief Write the results
 */
void Solver::save(GmshExport const &writer, std::string const &suffix)
{
    // Get data
    size_t nE = _pbl.getElements().size();
    size_t nM = _pbl.getNumModes();
    std::vector<sdpmDouble> omega = _pbl.getFrequencies();
    size_t nF = omega.size();
    sdpmDouble toDeg = 180. / 3.14159;

    // Write files
    std::cout << "Saving files... " << std::endl;
    // set up results
    Results results;
    // steady
    results.scalarsAtElems["mu"] = &_mu;
    results.scalarsAtElems["tau"] = &_tau;
    results.vectorsAtElems["u"] = &_u;
    results.scalarsAtElems["cp"] = &_cp;
    // unsteady
    std::vector<std::vector<std::vector<sdpmVector3d>>> u1_abs(nM, std::vector<std::vector<sdpmVector3d>>(nF, std::vector<sdpmVector3d>(nE)));
    std::vector<std::vector<std::vector<sdpmVector3d>>> u1_arg(nM, std::vector<std::vector<sdpmVector3d>>(nF, std::vector<sdpmVector3d>(nE)));
    std::vector<std::vector<std::vector<sdpmDouble>>> cp1_abs(nM, std::vector<std::vector<sdpmDouble>>(nF, std::vector<sdpmDouble>(nE)));
    std::vector<std::vector<std::vector<sdpmDouble>>> cp1_arg(nM, std::vector<std::vector<sdpmDouble>>(nF, std::vector<sdpmDouble>(nE)));
    for (size_t imd = 0; imd < nM; ++imd)
    {
        for (size_t ifq = 0; ifq < nF; ++ifq)
        {
            // set names
            std::stringstream u1_abs_name, u1_arg_name, cp1_abs_name, cp1_arg_name;
            u1_abs_name << "|u1|_" << imd << "_" << ifq;
            u1_arg_name << "< u1_" << imd << "_" << ifq;
            cp1_abs_name << "|cp1|_" << imd << "_" << ifq;
            cp1_arg_name << "< cp1_" << imd << "_" << ifq;
            // set values
            for (size_t i = 0; i < nE; ++i)
            {
                sdpmVector3cd u1 = _u1Re[imd][ifq][i] + sdpmComplex(0, 1) * _u1Im[imd][ifq][i];
                u1_abs[imd][ifq][i] = u1.array().abs();
                u1_arg[imd][ifq][i] = u1.array().arg() * toDeg;
                sdpmComplex cp1 = sdpmComplex(_cp1Re[imd][ifq][i], _cp1Im[imd][ifq][i]);
                cp1_abs[imd][ifq][i] = std::abs(cp1);
                cp1_arg[imd][ifq][i] = std::arg(cp1) * toDeg;
            }
            results.vectorsAtElems[u1_abs_name.str()] = &u1_abs[imd][ifq];
            results.vectorsAtElems[u1_arg_name.str()] = &u1_arg[imd][ifq];
            results.scalarsAtElems[cp1_abs_name.str()] = &cp1_abs[imd][ifq];
            results.scalarsAtElems[cp1_arg_name.str()] = &cp1_arg[imd][ifq];
        }
    }
    // save results
    writer.save(results, suffix);
    // save aerodynamic loads breakdown
    std::cout << "writing file: aeroloads_breakdown.dat... " << std::flush;
    std::ofstream outfile;
    outfile.open("aeroloads_breakdown.dat");
    for (auto b : _pbl.getBodies())
    {
        // steady
        outfile << "# Body - " << b->getName() << " (" << b->getElements().size() << " elements)" << std::endl;
        outfile << std::fixed << std::setprecision(5)
                << "Cl = " << b->getLiftCoeff() << "\n"
                << "Cd = " << b->getDragCoeff() << "\n"
                << "Cs = " << b->getSideforceCoeff() << "\n"
                << "Cm = " << b->getMomentCoeff() << "\n";
        // unsteady
        for (size_t imd = 0; imd < nM; ++imd)
        {
            for (size_t ifq = 0; ifq < nF; ++ifq)
            {
                sdpmComplex cl1 = sdpmComplex(b->getUnsteadyLiftCoeffReal(imd, ifq), b->getUnsteadyLiftCoeffImag(imd, ifq));
                sdpmComplex cd1 = sdpmComplex(b->getUnsteadyDragCoeffReal(imd, ifq), b->getUnsteadyDragCoeffImag(imd, ifq));
                sdpmComplex cs1 = sdpmComplex(b->getUnsteadySideforceCoeffReal(imd, ifq), b->getUnsteadySideforceCoeffImag(imd, ifq));
                sdpmComplex cm1 = sdpmComplex(b->getUnsteadyMomentCoeffReal(imd, ifq), b->getUnsteadyMomentCoeffImag(imd, ifq));
                outfile << "|Cl1| (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::abs(cl1) << "\n"
                        << "< Cl1 (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::arg(cl1) * toDeg << "\n"
                        << "|Cd1| (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::abs(cd1) << "\n"
                        << "< Cd1 (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::arg(cd1) * toDeg << "\n"
                        << "|Cs1| (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::abs(cs1) << "\n"
                        << "< Cs1 (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::arg(cs1) * toDeg << "\n"
                        << "|Cm1| (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::abs(cm1) << "\n"
                        << "< Cm1 (mode #" << imd << ", omega=" << omega[ifq] << ") = " << std::arg(cm1) * toDeg << "\n"
                        << std::endl;
            }
        }
    }
    outfile.close();
    std::cout << "done." << std::endl;
}

/**
 * @brief Compute steady flow functionals and loads on bodies
 */
void Solver::post(sdpmVectorXd const &etau, sdpmVectorXd const &emu)
{
    // Store source and doublet strength on elements
    std::vector<Body *> bodies = _pbl.getBodies();
    for (auto body : bodies)
    {
        for (auto e : body->getElements())
        {
            size_t id = e->getId() - 1;
            int i = _rows[e];
            _tau[id] = etau(i);
            _mu[id] = emu(i);
        }
    }

    // Transfer doublet at nodes
    std::vector<sdpmDouble> muNodes(_pbl.getNodes().size(), 0);
    for (auto b : bodies)
    {
        std::map<Node *, std::vector<Element *>> neMap = b->getMap();
        for (auto n : b->getNodes())
            for (auto e : neMap.at(n))
                muNodes[n->getId() - 1] += _mu[e->getId() - 1] / neMap.at(n).size();
    }

    // Compute surface flow functionals
    sdpmVector3d ui = _pbl.getDragDir(); // freestream velocity
    sdpmDouble mi = _pbl.getMach();      // freestream speed of sound
    for (auto body : bodies)
    {
        for (auto e : body->getElements())
        {
            sdpmVector3d u = ui - (_tau[e->getId() - 1] * e->getCompressibleNormal() + e->computeGradient(muNodes)).cwiseQuotient(sdpmVector3d(_pbl.getCompressibility(), 1., 1.));
            _u[e->getId() - 1] = u;                                                           // velocity
            _cp[e->getId() - 1] = 1 - (u.dot(u) - mi * mi * (u(0) - ui(0)) * (u(0) - ui(0))); // nonlinear pressure coefficient
        }
    }

    // Compute aerodynamic loads
    sdpmDouble sRef = _pbl.getRefArea();      // reference surface area
    sdpmDouble cRef = _pbl.getRefLgth();      // reference length
    sdpmVector3d xRef = _pbl.getRefCtr();     // reference center
    sdpmVector3d dragDir = _pbl.getDragDir(); // drag direction
    sdpmVector3d sideDir = _pbl.getSideDir(); // sideforce direction
    sdpmVector3d liftDir = _pbl.getLiftDir(); // lift direction
    _cl = 0;
    _cd = 0;
    _cs = 0;
    _cm = 0;
    for (auto b : _pbl.getBodies())
    {
        // Compute pressure loads coefficient (normalized by freestream dynamic pressure) on each element and average at node
        std::vector<Node *> nods = b->getNodes();
        std::map<Node *, std::vector<Element *>> neMap = b->getMap();
        std::vector<sdpmVector3d> nLoads(nods.size(), sdpmVector3d::Zero());
        for (size_t i = 0; i < nods.size(); ++i)
            for (auto e : neMap.at(nods[i]))
                nLoads[i] += -_cp[e->getId() - 1] * e->getSurfaceArea() * e->getNormal() / e->getNodes().size();
        b->setLoads(nLoads);

        // Compute integrated aerodynamic load coefficients (normalized by freestream dynamic pressure and reference area)
        // force coefficients along x, y and z directions and pitching moment around y axis
        sdpmDouble bcm = 0;
        sdpmVector3d cf(0, 0, 0);
        for (size_t i = 0; i < nods.size(); ++i)
        {
            sdpmVector3d l = nods[i]->getCoords() - xRef; // lever arm
            sdpmVector3d cfi = nLoads[i];                 // normalized force
            cf += cfi;
            bcm += (l(2) * cfi(0) - l(0) * cfi(2)) / (sRef * cRef); // moment is positive along y (3D) and negative along z (2D) => positive nose-up
        }
        b->setMomentCoeff(bcm);
        // rotate to flow direction
        b->setDragCoeff(cf.dot(dragDir) / sRef);
        b->setSideforceCoeff(cf.dot(sideDir) / sRef);
        b->setLiftCoeff(cf.dot(liftDir) / sRef);
        // compute total
        _cl += b->getLiftCoeff();
        _cd += b->getDragCoeff();
        _cs += b->getSideforceCoeff();
        _cm += bcm;
    }
}

/**
 * @brief Compute unsteady flow functionals and loads on bodies
 */
void Solver::post(size_t imd, size_t ifq, sdpmVectorXcd const &etau, sdpmVectorXcd const &emu)
{
    // Transfer doublet at nodes
    std::vector<sdpmDouble> muNodesRe(_pbl.getNodes().size(), 0), muNodesIm(_pbl.getNodes().size(), 0);
    for (auto b : _pbl.getBodies())
    {
        std::map<Node *, std::vector<Element *>> neMap = b->getMap();
        for (auto n : b->getNodes())
        {
            for (auto e : neMap.at(n))
            {
                muNodesRe[n->getId() - 1] += emu(_rows.at(e)).real() / neMap.at(n).size();
                muNodesIm[n->getId() - 1] += emu(_rows.at(e)).imag() / neMap.at(n).size();
            }
        }
    }

    // Compute flow and loads
    sdpmVector3d ui = _pbl.getDragDir(); // freestream velocity
    sdpmDouble mi = _pbl.getMach();
    sdpmDouble omega = _pbl.getFrequencies()[ifq];
    sdpmDouble sRef = _pbl.getRefArea();
    sdpmDouble cRef = _pbl.getRefLgth();
    sdpmVector3d xRef = _pbl.getRefCtr();
    sdpmDouble ca = cos(_pbl.getAos());
    sdpmDouble sa = sin(_pbl.getAos());
    _cl1Re[imd][ifq] = 0.;
    _cl1Im[imd][ifq] = 0.;
    _cd1Re[imd][ifq] = 0.;
    _cd1Im[imd][ifq] = 0.;
    _cs1Re[imd][ifq] = 0.;
    _cs1Im[imd][ifq] = 0.;
    _cm1Re[imd][ifq] = 0.;
    _cm1Im[imd][ifq] = 0.;
    for (auto b : _pbl.getBodies())
    {
        // compute velocity
        std::vector<sdpmVector3d> uc0 = b->getSteadyVelocity(imd);
        std::vector<sdpmVector3d> uc1 = b->getHarmonicVelocity(imd);
        std::vector<Element *> elems = b->getElements();
        std::vector<sdpmVector3cd> u1(elems.size());
        for (size_t i = 0; i < elems.size(); ++i)
        {
            u1[i] = uc0[i] + uc1[i] * sdpmComplex(0., 1.) * omega - (etau[_rows[elems[i]]] * elems[i]->getCompressibleNormal() + (elems[i]->computeGradient(muNodesRe) + sdpmComplex(0, 1) * elems[i]->computeGradient(muNodesIm))).cwiseQuotient(sdpmVector3d(_pbl.getCompressibility(), 1., 1.));
            _u1Re[imd][ifq][elems[i]->getId() - 1] = u1[i].real();
            _u1Im[imd][ifq][elems[i]->getId() - 1] = u1[i].imag();
        }

        // compute pressure
        for (size_t i = 0; i < elems.size(); ++i)
        {
            // phi_t, phi_x, phi_x(0), u^2
            sdpmComplex phit = -sdpmComplex(0., 1.) * omega * emu(_rows[elems[i]]);
            sdpmComplex phix = u1[i](0) - (uc0[i](0) + uc1[i](0) * sdpmComplex(0., 1.) * omega);
            sdpmComplex phix0(_u[elems[i]->getId() - 1](0) - ui(0), 0.);
            sdpmComplex cp1 = -sdpmComplex(2., 0.) * phit + sdpmComplex(2 * mi * mi, 0.) * (phix0 * phix + phix0 * phit);
            for (size_t j = 0; j < 3; ++j)
                cp1 -= sdpmComplex(2 * _u[elems[i]->getId() - 1](j), 0.) * u1[i](j);
            // cp
            _cp1Re[imd][ifq][elems[i]->getId() - 1] = cp1.real();
            _cp1Im[imd][ifq][elems[i]->getId() - 1] = cp1.imag();
        }

        // compute loads on nodes
        std::vector<Node *> nods = b->getNodes();
        std::map<Node *, std::vector<Element *>> neMap = b->getMap();
        std::vector<sdpmVector3cd> nLoads1(nods.size());
        for (size_t i = 0; i < nods.size(); ++i)
            for (auto e : neMap.at(nods[i]))
                nLoads1[i] += -sdpmComplex(_cp1Re[imd][ifq][e->getId() - 1], _cp1Im[imd][ifq][e->getId() - 1]) * e->getSurfaceArea() * e->getNormal() / e->getNodes().size();
        b->setUnsteadyLoads(imd, ifq, nLoads1);

        // compute moment coefficient and total load on body
        sdpmComplex bcm1(0., 0.);
        sdpmVector3cd cf1 = sdpmVector3cd::Zero();
        for (size_t i = 0; i < nods.size(); ++i)
        {
            sdpmVector3d l = nods[i]->getCoords() - xRef;
            cf1 += nLoads1[i];
            bcm1 += (l(2) * nLoads1[i](0) - l(0) * nLoads1[i](2)) / sdpmComplex(sRef * cRef, 0);
        }
        b->setUnsteadyMomentCoeff(imd, ifq, bcm1);

        // compute mean load coefficient
        sdpmVector3cd cf0 = sdpmVector3cd::Zero();
        std::vector<sdpmVector3d> loads = b->getLoads();
        for (size_t ii = 0; ii < loads.size(); ++ii)
            cf0 += loads[ii];
        // compute load coefficients, only valid for small AoAs
        b->setUnsteadyDragCoeff(imd, ifq, (cf1(0) * ca + cf1(1) * sa + (cf0(2) * b->getPitchAmplitude(imd) + cf1(2) * _pbl.getAoa()) * ca) / sRef);
        b->setUnsteadySideforceCoeff(imd, ifq, (-cf1(0) * sa + cf1(1) * ca - (cf0(2) * b->getPitchAmplitude(imd) + cf1(2) * _pbl.getAoa()) * sa) / sRef);
        b->setUnsteadyLiftCoeff(imd, ifq, (-(cf0(0) * b->getPitchAmplitude(imd) + cf1(0) * _pbl.getAoa()) + cf1(2)) / sRef);

        // compute total
        _cd1Re[imd][ifq] += b->getUnsteadyDragCoeffReal(imd, ifq);
        _cd1Im[imd][ifq] += b->getUnsteadyDragCoeffImag(imd, ifq);
        _cs1Re[imd][ifq] += b->getUnsteadySideforceCoeffReal(imd, ifq);
        _cs1Im[imd][ifq] += b->getUnsteadySideforceCoeffImag(imd, ifq);
        _cl1Re[imd][ifq] += b->getUnsteadyLiftCoeffReal(imd, ifq);
        _cl1Im[imd][ifq] += b->getUnsteadyLiftCoeffImag(imd, ifq);
        _cm1Re[imd][ifq] += b->getUnsteadyMomentCoeffReal(imd, ifq);
        _cm1Im[imd][ifq] += b->getUnsteadyMomentCoeffImag(imd, ifq);
    }
}

/**
 * @brief Compute Generalized Aerodynamic Forces matrix
 */
void Solver::computeGAF(size_t ifq)
{
    // Get sizes
    size_t n = _pbl.getNodes().size();
    size_t nm = _pbl.getNumModes();
    // Get x-component of reference center
    sdpmDouble xRef = _pbl.getRefCtr()(0);
    // Get z-component of displacements and nodal loads
    sdpmMatrixXd mz1 = sdpmMatrixXd::Zero(nm, n);
    sdpmMatrixXd lz1Re = sdpmMatrixXd::Zero(n, nm);
    sdpmMatrixXd lz1Im = sdpmMatrixXd::Zero(n, nm);
    for (size_t imd = 0; imd < nm; ++imd)
    {
        for (auto b : _pbl.getBodies())
        {
            sdpmDouble aAmp = b->getPitchAmplitude(imd);
            sdpmDouble hAmp = b->getPlungeAmplitude(imd);
            mz1.row(imd) = Eigen::Map<const sdpmVectorXd>(b->getModeZ(imd).data(), n).transpose();
            std::vector<Node *> nods = b->getNodes();
            std::vector<sdpmVector3d> l1Re = b->getUnsteadyLoadsReal(imd, ifq);
            std::vector<sdpmVector3d> l1Im = b->getUnsteadyLoadsImag(imd, ifq);
            for (size_t i = 0; i < nods.size(); ++i)
            {
                mz1(imd, nods[i]->getId() - 1) += aAmp * (nods[i]->getCoords()(0) - xRef) + hAmp;
                lz1Re(nods[i]->getId() - 1, imd) = -l1Re[i](2);
                lz1Im(nods[i]->getId() - 1, imd) = -l1Im[i](2);
            }
        }
    }
    // Compute GAF
    _q1Re[ifq] = mz1 * lz1Re;
    _q1Im[ifq] = mz1 * lz1Im;
}

/**
 * @brief Get real part of Generalized Aerodynamic Forces matrix at given frequency
 */
std::vector<std::vector<sdpmDouble>> Solver::getGafMatrixRe(size_t ifq) const
{
    return eigToStd(_q1Re[ifq]);
}

/**
 * @brief Get imaginary part of Generalized Aerodynamic Forces matrix at given frequency
 */
std::vector<std::vector<sdpmDouble>> Solver::getGafMatrixIm(size_t ifq) const
{
    return eigToStd(_q1Im[ifq]);
}

void Solver::write(std::ostream &out) const
{
    out << "sdpm::Solver with\n"
        << _pbl;
}
