# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# testing classes (for ctest)

ANSI_RED    = '\x1b[1;31m'
ANSI_GREEN  = '\x1b[1;32m'
ANSI_RESET  = '\x1b[0m'

class CTest:
    def __init__(self, name, val, expected, maxdiff=1e-10, refval=0.0, forceabs=False):
        self.name     = name                  # name of the test
        self.val      = float(val)            # calculated value
        self.expected = float(expected)       # expected value
        self.maxdiff  = abs(float(maxdiff))   # tolerance on max difference
        self.refval   = abs(float(refval))    # optional value used as denominator 
                                              #   if the expected one is close to 0
        self.forceabs = forceabs              # force the calculation of an absolute diff
        
    def run(self):
        ok = True
        adiff = abs(self.val-self.expected) # absolute diff
        
        # the ref value is the largest among the expected value and
        # the one provided by the user
        denom = max(abs(self.refval), abs(self.expected))
        
        if not self.forceabs and denom>self.maxdiff:
            diff = adiff / denom # relative diff
            typ='rel'
            percent = '%3.1f%%' % (self.maxdiff*100)
        else:
            diff = adiff # absolute diff
            typ='abs'
            percent = '%f' % self.maxdiff            
 
        print("[CTest] %s = %f (expected %f +/- %s)" % \
            (self.name, self.val, self.expected, percent)) 
                   
        if diff<=self.maxdiff:
            sgn = '<='
            info = "ok"
        else:
            sgn = '>'
            ok = False
            info = "wrong!"
        print("\t%s diff = %e %s %e [%s]" % (typ, diff, sgn, self.maxdiff, info))
        return ok   
            
class CTests:
    def __init__(self):
        self.tests = []
    def add(self, t):
        self.tests.append(t)
    def run(self):
        ok = True
        for i,t in enumerate(self.tests):
            #print "running test %d/%d..." % (i+1,len(self.tests))
            ok = t.run() and ok
            
        if ok:
            print(ANSI_GREEN + '** All tests are OK!' + ANSI_RESET)
        else:
            print('\n\n')
            raise Exception(ANSI_RED + '** Some tests failed!' + ANSI_RESET)
