# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# AGARD 445.6 wing (gradient)
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath, interpolate_modes
from sdpm.testing import *
import numpy as np

def main():
    # geometry parameters
    wnc = 20 # number of chordwise panels
    pars = {'nC': wnc, 'nS': 10, 'pC': 1.1, 'pS': 1.0}
    sref = 0.353 # reference area
    crd = 0.559 # root chord
    # flow and time parameters
    minf = 0.9 # freestream Mach number
    aoa = 1.0 * np.pi / 180 # angle of attack
    kred = 0.1 # reduced frequency
    n_modes = 4 # number of modes

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/agard445.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh)
    pbl.setGeometry(sref, crd, 0., 0., 0., True)
    pbl.setFreestream(aoa, 0., minf)
    pbl.setUnsteady(n_modes, [kred])
    wing = sdpm.Body(msh, 'wing', 'te', crd, wnc, n_modes, 1)
    x_modes, amp_modes = interpolate_modes(build_fpath('models/agard445_modes.csv', __file__, 1), wing.getNodes())
    for i in range(n_modes):
        wing.addMotion()
        wing.setFlexibleMotion(i, 1 / amp_modes[i], x_modes[:, 3*i:3*i+3])
    pbl.addBody(wing)
    tms['pbl'].stop()
    # solver
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    sol.update()
    sol.run()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # gradient evaluation
    tms['grd'].start()
    grd = sdpm.Gradient(sol)
    grd.run()
    tms['grd'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # recover GAF matrix from partial gradients (NB: since GAF is quadratic in DOF, result must be divided by 2)
    q_re = np.array(sol.getGafMatrixRe(0), dtype=float)
    q_im = np.array(sol.getGafMatrixIm(0), dtype=float)
    dq_re = np.zeros((n_modes, n_modes), dtype=float)
    dq_im = np.zeros((n_modes, n_modes), dtype=float)
    for i in range(n_modes):
        for j in range(n_modes):
            dq_re[i,j] = np.array(grd.computePartialsGafRe(0, j, i, j, 'dz'), dtype=float).dot(x_modes[:, 3*j+0] / amp_modes[j]) \
                       + np.array(grd.computePartialsGafRe(0, j, i, j, 'rx'), dtype=float).dot(x_modes[:, 3*j+1] / amp_modes[j]) \
                       + np.array(grd.computePartialsGafRe(0, j, i, j, 'ry'), dtype=float).dot(x_modes[:, 3*j+2] / amp_modes[j])
            dq_im[i,j] = np.array(grd.computePartialsGafIm(0, j, i, j, 'dz'), dtype=float).dot(x_modes[:, 3*j+0] / amp_modes[j]) \
                       + np.array(grd.computePartialsGafIm(0, j, i, j, 'rx'), dtype=float).dot(x_modes[:, 3*j+1] / amp_modes[j]) \
                       + np.array(grd.computePartialsGafIm(0, j, i, j, 'ry'), dtype=float).dot(x_modes[:, 3*j+2] / amp_modes[j])
            if i != j:
                dq_re[i,j] += np.array(grd.computePartialsGafRe(0, i, i, j, 'dz'), dtype=float).dot(x_modes[:, 3*i] / amp_modes[i])
                dq_im[i,j] += np.array(grd.computePartialsGafIm(0, i, i, j, 'dz'), dtype=float).dot(x_modes[:, 3*i] / amp_modes[i])
    dq_re *= 0.5
    dq_im *= 0.5

    # test
    tests = CTests()
    tests.add(CTest('Re(Q)',np.linalg.norm(q_re - dq_re), 0.0, 1e-10, forceabs=True))
    tests.add(CTest('Im(Q)',np.linalg.norm(q_im - dq_im), 0.0, 1e-10, forceabs=True))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
