# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# IO with Gmsh
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *

def main():
    # create timers
    tms = sdpm.Timers()
    # create and load mesh
    tms['msh_in'].start()
    pars = {'symY': 0, 'wNc': 2, 'wNs': 2}
    msh = MeshLoader(build_fpath('models/naca0012.geo', __file__, 1)).run(pars)
    tms['msh_in'].stop()
    # write mesh
    tms['msh_out'].start()
    wrt = sdpm.GmshExport(msh)
    wrt.save('_new')
    tst = sdpm.UnitTest()
    tst.gmshIO(msh, wrt)
    tms['msh_out'].stop()
    # print mesh info
    print('Mesh info:')
    print('-', msh)
    print('Nodes info:')
    for n in msh.getNodes():
        print('-', n)
    print('Elements info:')
    for e in msh.getElements():
        print('-', e)
    print('Tags info:')
    for n, t in msh.getTags().items():
        print('-', n, '->', t)
    print('')
    # print timers
    print(tms)

    # test
    tests = CTests()
    tests.add(CTest('Number of nodes', len(msh.getNodes()), 12, 1e-12))
    tests.add(CTest('Number of elements', len(msh.getElements()), 10, 1e-12))
    tests.add(CTest('Number of tags', len(msh.getTags()), 2, 1e-12))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
