# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def parse_args():
    """Parses command line arguments
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verb', help='increase output verbosity', action='count', default=0)
    parser.add_argument('--nogui', help='disable any graphical output', action='store_true')
    parser.add_argument('--clean', help='clean workspace', action='store_true')
    parser.add_argument('file', help='python files')
    args = parser.parse_args()
    return args

def build_fpath(name, caller, recurse_lvl=0):
    """Build absolute file path from a target and a caller

    Parameters
    ----------
    name : string
        name of the target file
    caller : string
        name of the file from which the target should be located
    recurse_lvl : int
        number of times to recurse to parent directory
    """
    import os.path
    recurse = ''
    for _ in range(recurse_lvl):
        recurse = os.path.join(recurse, '..')
    return os.path.abspath(os.path.join(os.path.dirname(caller), recurse, name))

def sort(conn, pts_vals):
    """Sort value array against line connectivity list

    Parameters
    ----------
    connct : 2d numpy array
        connectivity list (n_cells X n_point_by_cell)
    pts_vals : 2d numpy array
        points coordinates and values
    """
    import numpy as np
    # sort id vector
    conn = conn[conn[:,0].argsort(), :]
    # deep copy of data
    _copy = np.zeros((pts_vals.shape[0], pts_vals.shape[1]))
    _copy[:] = pts_vals
    # sort data against id
    _max = np.argmax(pts_vals[:, 0])
    pts_vals[0,:] = _copy[_max, :]
    nextId = conn[np.where(conn[:, 0]==_max)[0][0], 0]
    for i in range(1, conn.shape[0]):
        pts_vals[i,:] = _copy[conn[nextId,1], :]
        nextId = conn[nextId, 1]

def create_slices(body, sol, fname, ys):
    """Create slices along the span and extract the chordwise pressure coefficient distribution

    Parameters
    ----------
    body : sdpmBody
        body data structure containing grid points and cells
    sol : dict(string, array)
        name and solution data at cells
    fname : string
        name of the .vtu file (without extension)
    ys : array
        y-coordinates defining cutplanes
    """
    from sdpm.vtk_utils import write_grid, cut_grid, extract_data
    import numpy as np
    # Create grid and write to file
    grid = write_grid(fname, body, sol)
    # Extract data at along spanwise sections
    snames = list(sol.keys())
    for i in range(len(ys)):
        # get data on cut section
        coords, connct, cvalues = extract_data(cut_grid(grid, [0., ys[i], 0.]), snames)
        # normalize x and z-coordinates by local chord
        xc = np.zeros((coords.shape[0], 2))
        ile = np.argmin(coords[:,0]) # LE index
        crd = max(coords[:,0]) - min(coords[:,0]) # chord
        xc[:,0] = (coords[:,0] - coords[ile,0]) / crd
        xc[:,1] = (coords[:,2] - coords[ile,2]) / crd
        # average value at points from values at cell center
        nvalues = np.zeros((coords.shape[0], len(sol.keys())))
        for j in range(connct.shape[0]):
            pids = connct[j]
            for k in range(len(pids)):
                for l in range(len(sol.keys())):
                    nvalues[connct[j][k], l] += cvalues[snames[l]][j] / len(pids)
        # create dataset holding coordinates and values and sort it
        xc_vals = np.hstack((xc, nvalues))
        sort(connct, xc_vals)
        xc_vals = np.vstack((xc_vals, xc_vals[0,:]))
        # write to file
        hdr = f'y = {ys[i]}, c = {crd}\n'
        hdr += '{:>9s}, {:>10s}'.format('x/c', 'z/c')
        for name in snames:
            hdr += ', {:>10s}'.format(name)
        np.savetxt(fname+f'_slice_{i}.dat', xc_vals, fmt='%+1.4e', delimiter=',', header=hdr)

def interpolate_modes(fname, surf_nodes, flat=True):
    """Read modal displacements and interpolate them on surface grid

    Parameters
    ----------
    fname : string
        name of the CSV file containing the modes
    surf_nodes : array
        nodes of surface grid
    flat : bool
        whether inputs are provided on a flat surface (i.e. whether to use the z-coordinate)

    Return
    ------
    imodes : (n, 3*m) ndarray
        interpolated modes (n is the number of surf_nodes, m is the number of nodes)
    iamp : (m, 1) ndarray
        modal amplitudes
    """
    import numpy as np
    import scipy.interpolate as spip
    # Define number of columns containing coordinates to use
    ndim = 2 if flat else 3
    # Read modal data and remove duplicates
    modes = np.loadtxt(fname, delimiter=',', skiprows=1)
    _, idx = np.unique(modes[:,:ndim], axis=0, return_index=True)
    modes = modes[idx,:]
    # Interpolate on surface mesh
    xyz_surf = np.zeros((len(surf_nodes), ndim))
    for i, n in enumerate(surf_nodes):
        for j in range(ndim):
            xyz_surf[i,j] = n.getCoords()[j]
    imodes = spip.RBFInterpolator(modes[:,:ndim], modes[:,3:], smoothing=1)(xyz_surf)
    # Compute modes amplitude
    n_modes = (modes.shape[1] - 3) // 3
    iamp = np.zeros(n_modes)
    for i in range(n_modes):
        iamp[i] = np.linalg.norm(np.reshape(imodes[:, 3*i:3*i+3], (3 * len(surf_nodes), 1)))
    return imodes, iamp

def interpolate_pressure(fname, surf_nodes):
    """Read pressure derivative and interpolate it on surface grid

    Parameters
    ----------
    fname : string
        name of the CSV file containing the pressure derivative
    surf_nodes : array
        nodes of surface grid

    Return
    ------
    icp : (n, 1) ndarray
        interpolated pressure derivative (n is the number of surf_nodes)
    """
    import numpy as np
    import scipy.interpolate as spip
    # Read pressure and remove duplicates
    cp = np.loadtxt(fname, delimiter=',', skiprows=1)
    _, idx = np.unique(cp[:,:3], axis=0, return_index=True)
    cp = cp[idx,:]
    # Interpolate on surface mesh
    xyz_surf = np.zeros((len(surf_nodes), 3))
    for i, n in enumerate(surf_nodes):
        for j in range(3):
            xyz_surf[i,j] = n.getCoords()[j]
    icp = spip.RBFInterpolator(cp[:,:3], cp[:,3:], smoothing=1e-6)(xyz_surf)
    return icp
