# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import vtk

def read_grid(fname):
    """Read VTU file from disk

    Parameters
    ----------
    fname : string
        name of the .vtu file (without extension)

    Return
    ------
    grid : vtkUnstructuredGrid
        grid in VTU format containing points, cells, and cells data
    """
    import os.path
    # Check
    fname += '.vtu'
    if not os.path.exists(fname):
        raise RuntimeError('File does not exists!\n')
    # Read
    reader = vtk.vtkXMLUnstructuredGridReader()
    reader.SetFileName(fname)
    reader.Update()
    return reader.GetOutputPort()

def write_grid(fname, body, sol):
    """Write VTU file from disk

    Parameters
    ----------
    fname : string
        name of the .vtu file (without extension)
    body : sdpmBody
        body data structure containing grid points and cells
    sol : dict(string, array)
        name and solution data at cells

    Return
    ------
    grid : vtkUnstructuredGrid
        grid in VTU format containing points, cells, and cells data
    """
    # Create grid data structure
    grid = vtk.vtkUnstructuredGrid()
    points = vtk.vtkPoints()
    cells = vtk.vtkCellArray()
    ctypes = []

    # Add points
    node_point_map = {}
    for n in body.getNodes():
        coords = n.getCoords()
        node_point_map[n.getId()] = points.InsertNextPoint(coords[0], coords[1], coords[2])
    grid.SetPoints(points)

    # Add cells
    elem_cell_map = {}
    for e in body.getElements():
        # create cell and set number of points
        nodes = e.getNodes()
        cell = vtk.vtkQuad()
        cell.GetPointIds().SetNumberOfIds(len(nodes))
        # add each vertex point
        for i, n in enumerate(nodes):
            cell.GetPointIds().SetId(i, node_point_map[n.getId()])
        # add cell to grid
        elem_cell_map[e.getId()] = cells.InsertNextCell(cell)
        ctypes.append(cell.GetCellType())
    grid.SetCells(ctypes, cells)

    # Add values at cells
    for name, vals in sol.items():
        # fill the values
        values = vtk.vtkDoubleArray()
        for e in body.getElements():
            eid = e.getId()
            values.InsertValue(elem_cell_map[eid], vals[eid - 1])
        # add values to the grid
        values.SetName(name)
        grid.GetCellData().AddArray(values)

    # Write to file
    fname += '.vtu'
    writer = vtk.vtkXMLUnstructuredGridWriter()
    writer.SetFileName(fname)
    writer.SetInputData(grid)
    writer.Write()
    return grid

def cut_grid(grid, org, nrm=[0.,1.,0.]):
    """Perform a cut on a grid using a plane defined by its origin and normal vector

    Parameters
    ----------
    grid : vtkUnstructuredGrid
        grid in VTU format containing points, cells, and cells data
    org : array
        origin of cut plane
    nrm : array
        normal vector to cut plane

    Return
    ------
    vtkPolyData
        dataset holding points, cells and data
    """
    # Create cutplane
    plane = vtk.vtkPlane()
    plane.SetOrigin(org[0], org[1], org[2])
    plane.SetNormal(nrm[0], nrm[1], nrm[2])
    # Create cutting tool
    cutter = vtk.vtkCutter()
    cutter.SetCutFunction(plane)
    cutter.SetInputData(grid)
    # Perform cut
    cutter.Update()
    return cutter.GetOutput()

def extract_data(data, vname):
    """Extract point coordinates, connectivity list and values from a dataset

    Parameters
    ----------
    data : vtkPolyData
        dataset holding points, cells and data

    Return
    ------
    coords : 2d numpy array
        array of coordinates (n_points X 3)
    connct : 2d numpy array
        connectivity list (n_cells X n_point_by_cell)
    values : dict(name, 2d numpy array)
        name and data at cut cells
    """
    import numpy as np
    # Transfer point coordinates
    _pts = data.GetPoints()
    pts = np.zeros((_pts.GetNumberOfPoints(), 3))
    for i in range(pts.shape[0]):
        for j in range(3):
            pts[i][j] = _pts.GetPoint(i)[j]
    # Transfer connectivity
    _elems = data.GetLines().GetData()
    nvert = 2
    elems = np.zeros((_elems.GetNumberOfTuples() // (nvert+1), nvert), dtype=int)
    for i in range(elems.shape[0]):
            for j in range(nvert):
                elems[i][j] = _elems.GetTuple((nvert+1)*i+j+1)[0]
    # Transfer variables
    vals = {}
    for name in vname:
        _vals =  data.GetCellData().GetArray(name)
        vals[name] = np.zeros((_vals.GetNumberOfTuples(), _vals.GetNumberOfComponents()))
        for i in range(vals[name].shape[0]):
            for j in range(vals[name].shape[1]):
                vals[name][i][j] = _vals.GetTuple(i)[j]
    return pts, elems, vals
